# Copyright (C) 2008 Red Hat, Inc

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# a long with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Copyright (C) 2008 Red Hat, Inc

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# a long with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
require 'uri'
require 'git'
require 'forwardable'

# It might be good to put this somewhere where it's reusable
module ClassAccessorsWithDefaults
  module ClassMethods
    @@_outer = self

    def cattr_accessor(name, default=nil)
      @@_outer.instance_eval { attr_accessor name }
      instance_variable_set("@" + name.to_s, default)
    end
  end

  def self.included(klass)
    klass.extend(ClassMethods)
  end

  def method_missing(meth, *args)
    self.class.respond_to?(meth) ? self.class.send(meth, *args) : super(meth, *args)
  end
end

module GenomeSync
  # TODO: the client shouldn't have to know about @logger!
  module Loggable
    def log(msg)
      @logger.info(msg) unless @logger.nil?
    end
  end

  class Syncronizer
    include ClassAccessorsWithDefaults
    include Loggable
    extend Forwardable

    def_delegator :@saver, :save

    cattr_accessor :localcache_remote_name, "localcache"
    attr_reader :working_path, :remote

    def initialize(remote_url, local_path, working_dir, logger=nil)
      # The local git-dir
      @local_path = local_path

      # Where we clone from
      @remote_url = remote_url
      @remote = URI.parse(remote_url).host

      @working_dir = working_dir
      @working_path = File.join(@working_dir, @local_path)

      @logger = logger

      @saver = Saver.new(@local_path, @working_path)

      # This could easily be done with reflection
      log("local_path = #{@local_path}")
      log("remote_url = #{@remote_url}")
      log("remote = #{@remote}")
      log("working_dir = #{@working_dir}")
      log("working_path = #{@working_path}")
    end

    # Sets up a git repo whose local branches reflect the state of the
    # @local_path.  A remote is setup and fetched based on the remote_url.
    # 
    # If passed a block yields the git repo object
    def start
      setup
      git = safe_open
      setup_remotes(git)

      # Don't bother asking the user anything else about this repo if we
      # don't even have a local public copy yet.
      if block_given? && File.directory?(@local_path)
        yield git 
      else 
        log("Skipping questions for #{@local_path}")
      end

      return true
    end

    def set_local_branches_to_remote(git, remote)
      remote_branches = git.branches.remote.find_all do |b| 
        next if b.name == 'HEAD'
        b.remote.name == remote 
      end
 
      remote_branches.each {|b| checkout_b(git, b)}
    end

    private 

    def setup
      log "making #{@working_dir}"
      FileUtils.mkdir_p(@working_dir)

      make_repo_parent_dir
    end

    # I couldn't find a way to 'checkout -b --force' so this will have to do for
    # now.  It would be better to work this into the git gem at some point.
    def checkout_b(git, branch, opts = {})
      log("Setting #{branch.name} to #{git.object(branch).sha}")
      git.branch(branch.name).checkout
      git.reset_hard(branch) 
    end

    def safe_open
      log "@working_path = #{@working_path}"

      # We need to clone from the remote if the local clone doesn't exist
      if File.directory?(@working_path)
        log "Opening #{@working_path}"
        return Git.open(@working_path) 
      else
        # NOTE: For symplicity's sake I decided not to bother cloning from /srv/git
        # locally if the repo does not exist in the working dir.  This would
        # make the tool have to detect if the local clone is out sync.  While
        # this isn't hard, I don't feel like messing with it at the moment.
        log("Cloning #{@remote_url} to #{@working_path}")
        g = Git.clone(@remote_url, @working_path)

        set_local_branches_to_remote(g, "origin")

        return g 
      end
    end

    def setup_remotes(git)
      # Don't add the remote if it already exists
      # Note: By convention the remote must be the same as the genome repo fqdn
      git.add_remote(@remote, @remote_url) unless git.remotes.find {|r| r.name == @remote}
      log("Fetching #{@remote}")
      git.fetch(@remote)

      # If we have a public repo make sure we have a remote that points to it
      if File.directory?(@local_path) # This dir will exist if the user has ever used 'save'
        # Add a remote if we don't have one
        unless git.remotes.find {|r| r.name == localcache_remote_name}
          git.add_remote(self.localcache_remote_name, @local_path) 
        end

        log("Fetching #{localcache_remote_name}")
        git.fetch(localcache_remote_name)

        # reset hard all our local branches
        set_local_branches_to_remote(git, localcache_remote_name)
      end
    end

    def make_repo_parent_dir
      repo_parent_dir = File.dirname(@working_path) 
      log "Making #{repo_parent_dir}"
      FileUtils.mkdir_p(repo_parent_dir)
    end
  end

  class Saver
    include ClassAccessorsWithDefaults
    include Loggable

    cattr_accessor :puppetmaster_base_dir, "/var/lib/puppet/modules/genome"
    cattr_accessor :puppetmaster_tracking_branch, "master"

    def initialize(local_path, working_path)
      # The local git-dir
      @local_path = local_path
      @local_repo_parent_dir, @local_repo_name = File.split(@local_path)

      @working_path = working_path
    end

    def save
      unless File.directory?(@local_path)
        log("Making #{@local_repo_parent_dir}")
        FileUtils.mkdir_p(@local_repo_parent_dir)
         
        do_local_clone
      end

      push_content
    end

    private

    def push_content
      git = Git.open(@working_path)

      log("Pushing all branches in #{@working_path} to #{@local_path}:")
      git.branches.local.each do |b| 
        next if b.name == "(no branch)" #important for superprojects
        b.checkout
        log("#{b}")
        git.push(@local_path, b, :force => true)
      end
      log("done.")
    end

    def do_local_clone
      log("Cloning #{@working_path} to #{@local_path}")
      # Handle the special puppet module case
      if @local_repo_parent_dir =~ %r-puppet- 
        # For the correct code to make it to the puppetmaster we need to
        # make sure we checkout the desired tracking branch.
        Git.open(@working_path).checkout(puppetmaster_tracking_branch)

        puppetmaster_dir = File.join(puppetmaster_base_dir, @local_repo_name)
        log("Checking out 'master' at #{puppetmaster_dir}")
        Git.clone(@working_path, puppetmaster_dir)
        # HACK: I couldn't get the git gem to clone the working dir for the
        # puppetmaster and the bare repo for gitweb.  This is a workaround.
        FileUtils.mv(File.join(puppetmaster_dir, ".git"), @local_path)
      else 
        Git.clone(@working_path, 
                  @local_repo_name, 
                  :path => @local_repo_parent_dir,
                  :bare => true)
      end
    end
  end
end
