# Copyright (C) 2008 Red Hat, Inc

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# a long with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# These are required by the Rakefile.  The tests should probably run in a separate process.
$:.unshift(File.dirname(__FILE__) + "/../lib")
require 'genome-sync/util'
require 'rubygems'
require 'git'
require 'test/unit'
#require 'mocha'

class GenomeSyncTestCase < Test::Unit::TestCase
  def initialize(*args)
    super *args
    @tmpdir = File.dirname(__FILE__) + "/tmp"
    @tmprepo_path = @tmpdir + "/upstream_repo"
    @tmprepo_uri = "file://localhost" + @tmprepo_path
    @tmpworking_dir = @tmpdir + "/working"
    @tmpfile_basename = "tmpfile"
    @test_local_path = @tmpdir + "/local_repo"
  end

  def teardown
    FileUtils.rm_rf(@tmpdir)
  end

  def create_tmp_git(repo_path)
    FileUtils.mkdir_p(repo_path)
    git = Git.init(repo_path)

    FileUtils.cd(repo_path) do
      # We need at least one commit
      FileUtils.touch(@tmpfile_basename)
      git.add(File.basename(@tmpfile_basename))
      git.commit("Committing the #{@tmpfile_basename}")

      # We need a few branches as well
      4.times {|i| git.branch("branch#{i}").create}
    end
  end

  # Test::Unit does have support for abstract tests.  This functionality could
  # always be moved to a module.
  def test_this_is_stupid; end
end

class TestSync < GenomeSyncTestCase
  include GenomeSync

  def setup
    create_tmp_git(@tmprepo_path)
  end

  def test_remote_parse
    s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    assert_equal("localhost", s.remote)
  end

  def test_working_path
    s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    assert_equal(@tmpworking_dir + @test_local_path, s.working_path)
  end

  def test_required_dirs_are_created
    s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    s.start

    assert(File.directory?(@tmpworking_dir))
    assert(File.directory?(File.dirname(s.working_path)))
  end

  def test_start_is_reentrant
    assert_nothing_raised do
      Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir).start 
      Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir).start 
      Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir).start
    end
  end

  def test_remote_branches_become_local
    executed_block = false

    s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    s.start
    s.save # creates @test_local_path which is needed in order for the block to fire

    Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir).start do |git|
      executed_block = true

      local_branches = git.branches.local
      assert(local_branches.size > 1)

      remote_branches = git.branches.remote.find_all do |b| 
        b.remote.name == URI.parse(@tmprepo_uri).host
      end

      assert_equal(remote_branches.size, local_branches.size)
      
      local_branches.each do |l| 
        assert_not_nil(remote_branches.find {|r| r.name == l.name })
      end
    end

    assert(executed_block)
  end

  def test_if_local_repo_exists_we_must_reset_hard_to_its_state
    # Setup the local_path
    s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    s.start
    s.save

    # Create a new clone to work with
    clone_dir = "tmprepo_clone"
    git1 = Git.clone(@tmprepo_uri, clone_dir, :path => @tmpworking_dir)

    # Get the local_path master branch out of sync with the working_dir
    newfile_name = "newfile"
    FileUtils.touch(File.join(@tmpworking_dir, clone_dir, newfile_name))
    git1.add(newfile_name)
    git1.commit("Adding a new commit")
    git1.push(@test_local_path)
    log1 = git1.log
    assert(log1.size > 1)

    # For the working_path to be updated
    s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    s.start
    s.save

    # Make sure it worked
    git2 = Git.open(s.working_path)
    git2.checkout("master")
    assert_equal(log1.to_s, git2.log.to_s)
  end

  def test_save_method
    s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    s.start
    s.save
    assert_nothing_raised { Git.bare(@test_local_path) }
  end

  # This is to avoid asking the user excess questions
  def test_yield_not_called_if_local_path_doesnt_exist
    s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    s.start do
      flunk "If you made it here you FAIL"
    end
  end
end

class TestSaver < GenomeSyncTestCase
  include GenomeSync

  def setup
    create_tmp_git(@tmprepo_path)
    @s = Syncronizer.new(@tmprepo_uri, @test_local_path, @tmpworking_dir)
    @s.start
  end

  def test_save_pushes_all_branches
    Saver.new(@test_local_path, @s.working_path).save
    assert(File.directory?(@test_local_path))

    working_git = Git.open(@s.working_path)
    saved_git = Git.bare(@test_local_path)

    assert(working_git.branches.local.size > 1)
    working_git.branches.local.each do |l| 
      assert_not_nil(saved_git.branches.find {|s| s.name == l.name })
    end
  end
end

class TestSaverPuppet < GenomeSyncTestCase
  include GenomeSync

  def initialize(*args)
    super *args
    @tmppuppet_base = @tmpdir + "/puppetbase"
    @tmppuppet_repo_path = @tmpdir + "/upstream_puppet_repo"
    @tmppuppet_repo_uri = "file://localhost" + @tmppuppet_repo_path
    @puppet_test_local_path = File.join(@test_local_path, "/puppet-modules/module_foo")
  end

  def setup
    Saver.puppetmaster_base_dir = @tmppuppet_base

    create_tmp_git(@tmppuppet_repo_path)
    @s = Syncronizer.new(@tmppuppet_repo_uri, @puppet_test_local_path, @tmpworking_dir)
    @s.start
  end

  def test_pushing_puppet_module_yields_disconnected_repo
    Saver.new(@puppet_test_local_path, @s.working_path).save
    assert_nothing_raised { Git.bare(@puppet_test_local_path) }

    expected_file = File.join(@tmppuppet_base, File.basename(@puppet_test_local_path), @tmpfile_basename)
    assert(File.exist?(expected_file))
  end

  def test_puppetmaster_working_dir_is_on_tracking_branch
    Saver.new(@puppet_test_local_path, @s.working_path).save

    puppetmaster_working_dir = File.join(Saver.puppetmaster_base_dir, File.basename(@puppet_test_local_path))

    # the repository option should be called git_dir
    actual_branch = Git.open(puppetmaster_working_dir, :repository => @puppet_test_local_path).branch.name
    assert_equal(Saver.puppetmaster_tracking_branch, actual_branch)
  end
end

class TestClassAccessorsWithDefaults < Test::Unit::TestCase
  def setup
    @foo_klass = Class.new do
      include ClassAccessorsWithDefaults
      cattr_accessor :bar, "buz"
    end
    
    @f = @foo_klass.new
  end

  def test_class_accessor_become_instance_readers
    assert_equal("buz", @foo_klass.bar)
    assert_equal("buz", @f.bar)
  end

  def test_exception_thrown_for_bad_cattr
    assert_raise(NoMethodError) { @f.opps }
  end
end
